#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <gnutls/gnutls.h>

#ifndef _WIN32
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
#else
 #include <conio.h>
 #include <winsock2.h>
 typedef int socklen_t;
 #define close closesocket
#endif

#define write_uint32le(dest, src) do {\
    (dest)[0] = ((src)&0x000000FF)>>(8*0); \
    (dest)[1] = ((src)&0x0000FF00)>>(8*1); \
    (dest)[2] = ((src)&0x00FF0000)>>(8*2); \
    (dest)[3] = ((src)&0xFF000000)>>(8*3); \
} while(0)
#define write_uint32be(dest, src) do {\
    (dest)[0] = ((src)&0xFF000000)>>(8*3); \
    (dest)[1] = ((src)&0x00FF0000)>>(8*2); \
    (dest)[2] = ((src)&0x0000FF00)>>(8*1); \
    (dest)[3] = ((src)&0x000000FF)>>(8*0); \
} while(0)
#define write_uint16be(dest, src) do {\
    (dest)[0] = ((src)&0xFF00)>>(8*1); \
    (dest)[1] = ((src)&0x00FF)>>(8*0); \
} while(0)
#define write_uint8be(dest, src) do {\
    (dest)[0] = ((src)&0xFF)>>(8*0); \
} while(0)

static gnutls_session_t tls_session = 0;
static gnutls_dh_params_t dh_params = 0;
static gnutls_certificate_credentials_t x509_cred = 0;
static gnutls_priority_t priority_cache = 0;
static FILE * logfd = NULL;
static int listenfd = 0, sessionfd = 0;

#ifdef _WIN32
static HANDLE EventObjects[2];
static HANDLE StdinThread = NULL;
static char line[4096];
static volatile LONG NewCommand = false;

static DWORD WINAPI StdinProc(void *){
    while(1){
        if(fgets(line, 4096, stdin) != NULL){
            InterlockedExchange(&NewCommand, true);
            SetEvent(EventObjects[0]);
        }
        WaitForSingleObject(EventObjects[1], INFINITE);
    }
    return 0;
}
#endif

static inline void writetolog(const char *msg, size_t size){
    fwrite(msg, 1, size, logfd);
    fflush(logfd);
}

static void Shutdown(){
    if(StdinThread) TerminateThread(StdinThread, 0);

    gnutls_deinit(tls_session);
    gnutls_priority_deinit(priority_cache);
    gnutls_certificate_free_credentials(x509_cred);
    gnutls_dh_params_deinit(dh_params);
    gnutls_global_deinit();

    close(sessionfd);
    close(listenfd);
    if(logfd) fclose(logfd);

    #ifdef _WIN32
    WSACleanup();
    #endif
}

static void Shutdown_M(const char * Message){
    fprintf(stderr, "Error: %s (%s).\n", Message, strerror(errno));
    Shutdown();
    exit(EXIT_FAILURE);
}

int main(int, char **)
{
    int ret;

    #ifdef _WIN32
    WSADATA wsaData;
    WSAStartup(MAKEWORD(2, 2), &wsaData);
    EventObjects[0] = CreateEvent(NULL, FALSE, FALSE, NULL);
    EventObjects[1] = CreateEvent(NULL, FALSE, FALSE, NULL);
    StdinThread = CreateThread(NULL, 0, StdinProc, NULL, 0, NULL);
    if(StdinThread == NULL)
        Shutdown_M("Failed to create stdin thread");
    #endif

    logfd = fopen("log.dat", "wb");
    if(!logfd)
        Shutdown_M("Could not open log.dat for writing");

    if(gnutls_global_init() < 0)
        Shutdown_M("Failed to initialize GnuTLS");

    gnutls_dh_params_init(&dh_params);
    gnutls_dh_params_generate2(dh_params, 1024);
    gnutls_certificate_allocate_credentials(&x509_cred);
    gnutls_certificate_set_x509_key_file(x509_cred, "cert.pem", "key.pem", GNUTLS_X509_FMT_PEM);
    gnutls_certificate_set_dh_params(x509_cred, dh_params);
    gnutls_priority_init(&priority_cache, "NORMAL", NULL);

    if((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        Shutdown_M("Could not open socket");

    struct sockaddr_in server_addr;
    memset(&server_addr, 0, sizeof(sockaddr_in));
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons(49100);
    server_addr.sin_addr.s_addr = INADDR_ANY;

    if(bind(listenfd, (struct sockaddr *) &server_addr, sizeof(server_addr)) < 0)
        Shutdown_M("Could not bind socket");

    if(listen(listenfd, 5) < 0)
        Shutdown_M("Could not listen on socket");

    while(1){
        printf("Waiting for connections...\n");

        struct sockaddr_in client_addr;
        socklen_t client_len = sizeof(client_addr);
        if((sessionfd = accept(listenfd, (struct sockaddr *) &client_addr, &client_len)) < 0)
            continue;

        printf("Accepted connection.\n");
        writetolog("\n\n[cityserver: New connection]\n", sizeof("\n\n[cityserver: New connection]\n")-1);

        if(gnutls_init(&tls_session, GNUTLS_SERVER) < 0)
            Shutdown_M("Failed to open a GnuTLS session");

        gnutls_priority_set(tls_session, priority_cache);
        gnutls_credentials_set(tls_session, GNUTLS_CRD_CERTIFICATE, x509_cred);
        gnutls_certificate_server_set_request(tls_session, GNUTLS_CERT_REQUEST);

        gnutls_transport_set_ptr(tls_session, (gnutls_transport_ptr_t) sessionfd);
        if((ret = gnutls_handshake(tls_session)) < 0){
            printf("TLS handshake failed (%s).\n", gnutls_strerror(ret));
            gnutls_deinit(tls_session); tls_session = 0;
            close(sessionfd); sessionfd = 0;
            continue;
        }

        #ifdef _WIN32
        if(WSAEventSelect(sessionfd, EventObjects[0], FD_READ|FD_OOB|FD_CLOSE) != 0)
            Shutdown_M("WSAEventSelect failed");
        #endif

        line[0] = '\0';
        printf("To see the list of commands, type 'help'.\n");

        uint32_t usercommand = 1;
        bool showcommand = true;

        while(1){
            if(showcommand){
                printf("\ncityserver:%u> ", usercommand);
                showcommand = false;
            }

            #ifdef _WIN32
            if(WaitForSingleObject(EventObjects[0], INFINITE) != 0)
                Shutdown_M("WaitForSingleObject failed");

            if(NewCommand){
                NewCommand = false;

                char command[33] = "";
                if(sscanf(line, "%32s", command) > 0 && command[0] != '\0' && command[0] != ' '
                    && command[0] != '\r' && command[0] != '\n'){
                    if(!strcmp(command, "help")){
                        printf("Commands:\n"
                            "    help:          Show this help screen\n"
                            "    send:          Send a packet\n"
                            "                   arg1: Name of file to read packet data from\n"
                            "    close:         Close the connection with the client\n");
                    }else if(!strcmp(command, "send")){
                        char filename[256];
                        if(sscanf(line, "%32s %256s", command, filename) != 2)
                            printf("Invalid arguments");
                        else{
                            FILE *fd = fopen(filename, "rb");
                            if(!fd)
                                printf("Failed to open file '%s' for reading (%s)", line, strerror(errno));
                            else{
                                size_t size = fread(line, 1, 4096, fd);
                                fclose(fd);
                                if(size == 0)
                                    printf("Failed to read from file '%s' (%s)", line, strerror(errno));
                                else{
                                    gnutls_record_send(tls_session, line, size);
                                    fprintf(logfd, "\nSent %u bytes:", size);
                                    writetolog(line, size);
                                    printf("    ==> Sent %u bytes.", size);
                                }
                            }
                        }
                    }else if(!strcmp(command, "close")){
                        printf("    *** Closed connection to client.\n\n");
                        line[0] = '\0';
                        SetEvent(EventObjects[1]);
                        break;
                    }else printf("Unrecognized command.");
                }

                line[0] = '\0';
                SetEvent(EventObjects[1]);
                showcommand = true;
                usercommand++;
            }

            char buffer[4096];
            ssize_t size;
            while((size = gnutls_record_recv(tls_session, buffer, 4096)) > 0){
                fprintf(logfd, "\nReceived %u bytes:\n", size);
                writetolog(buffer, size);
                printf("\n    <== Received %u bytes.", size);
                showcommand = true;
            }
            if(size != GNUTLS_E_AGAIN){
                if(size == 0) printf("\n    *** Client disconnected.\n\n");
                else printf("\n    *** Lost connection to client (%s).\n\n", gnutls_strerror(size));
                break;
            }
            #endif
        }

        gnutls_deinit(tls_session); tls_session = 0;
        close(sessionfd); sessionfd = 0;
    }

    Shutdown();
    printf("Shut down gracefully.\n");

    return 0;
}